extern crate core;

#[cfg(test)]
pub mod tests {
    use flare_kernel::{Coefs, TransformSpec, VariationKind, VariationParams, VariationSpec};
    use futures::channel::oneshot;
    use futures_executor::block_on;
    use std::default::Default;
    use wgpu::util::DeviceExt;

    async fn create_instance() -> wgpu::Instance {
        let backends = wgpu::util::backend_bits_from_env().unwrap_or_default();
        wgpu::Instance::new(wgpu::InstanceDescriptor {
            backends,
            ..Default::default()
        })
    }

    async fn create_adapter(instance: &wgpu::Instance) -> wgpu::Adapter {
        instance
            .request_adapter(&wgpu::RequestAdapterOptions {
                power_preference: wgpu::PowerPreference::HighPerformance,
                force_fallback_adapter: false,
                compatible_surface: None,
            })
            .await
            .expect("Failed to create adapter")
    }

    async fn create_device_queue(adapter: &wgpu::Adapter) -> (wgpu::Device, wgpu::Queue) {
        adapter
            .request_device(
                &wgpu::DeviceDescriptor {
                    label: Some("unit test"),
                    required_features: wgpu::Features::empty(),
                    required_limits: wgpu::Limits::default(),
                    memory_hints: wgpu::MemoryHints::default(),
                },
                None,
            )
            .await
            .expect("Failed to request device queue")
    }

    async fn create_shader(device: &wgpu::Device) -> wgpu::ShaderModule {
        device.create_shader_module(wgpu::include_spirv!(concat!(
            env!("OUT_DIR"),
            "/shader_binary.spv"
        )))
    }

    async fn create_pipeline_layout(
        device: &wgpu::Device,
        layout: &wgpu::BindGroupLayout,
    ) -> wgpu::PipelineLayout {
        device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
            label: Some("pipeline"),
            bind_group_layouts: &[layout],
            push_constant_ranges: &[],
        })
    }

    async fn create_apply_transform_f32_bind_group_layout(
        device: &wgpu::Device,
    ) -> wgpu::BindGroupLayout {
        device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
            label: Some("apply_transform_f32"),
            entries: &[
                wgpu::BindGroupLayoutEntry {
                    binding: 0,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                wgpu::BindGroupLayoutEntry {
                    binding: 1,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                wgpu::BindGroupLayoutEntry {
                    binding: 2,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Uniform,
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                wgpu::BindGroupLayoutEntry {
                    binding: 3,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: false },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
            ],
        })
    }

    async fn create_apply_transform_f32_compute_pipeline(
        device: &wgpu::Device,
        pipeline_layout: &wgpu::PipelineLayout,
        shader: &wgpu::ShaderModule,
    ) -> wgpu::ComputePipeline {
        device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
            label: Some("apply_transform_f32_compute_pipeline"),
            layout: Some(pipeline_layout),
            module: shader,
            entry_point: Some("apply_transform_f32"),
            compilation_options: Default::default(),
            cache: Default::default(),
        })
    }

    async fn do_apply_transform_f32(
        transform: &TransformSpec,
        variations: &[VariationSpec],
        point: &glam::Vec2,
    ) -> glam::Vec2 {
        let instance = create_instance().await;
        let adapter = create_adapter(&instance).await;
        let (device, queue) = create_device_queue(&adapter).await;

        let shader = create_shader(&device).await;
        let bind_group_layout = create_apply_transform_f32_bind_group_layout(&device).await;
        let pipeline_layout = create_pipeline_layout(&device, &bind_group_layout).await;
        let compute_pipeline =
            create_apply_transform_f32_compute_pipeline(&device, &pipeline_layout, &shader).await;

        let transform_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("transform buffer"),
            usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_DST,
            contents: bytemuck::cast_slice(&[*transform]),
        });

        let variation_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("variation buffer"),
            usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_DST,
            contents: bytemuck::cast_slice(variations),
        });

        let point_in_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("point in buffer"),
            usage: wgpu::BufferUsages::UNIFORM | wgpu::BufferUsages::COPY_DST,
            contents: bytemuck::cast_slice(&[*point]),
        });

        let point_out_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("point out buffer"),
            size: size_of::<glam::Vec2>() as u64,
            usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_SRC,
            mapped_at_creation: false,
        });

        let point_out_staging = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("point out staging buffer"),
            size: size_of::<glam::Vec2>() as u64,
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
            label: None,
            layout: &bind_group_layout,
            entries: &[
                wgpu::BindGroupEntry {
                    binding: 0,
                    resource: transform_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 1,
                    resource: variation_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 2,
                    resource: point_in_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 3,
                    resource: point_out_buffer.as_entire_binding(),
                },
            ],
        });

        let mut encoder =
            device.create_command_encoder(&wgpu::CommandEncoderDescriptor { label: None });

        {
            let mut compute_pass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor {
                label: None,
                timestamp_writes: None,
            });

            compute_pass.set_pipeline(&compute_pipeline);
            compute_pass.set_bind_group(0, &bind_group, &[]);

            compute_pass.dispatch_workgroups(1, 1, 1);
        }

        encoder.copy_buffer_to_buffer(
            &point_out_buffer,
            0,
            &point_out_staging,
            0,
            size_of::<glam::Vec2>() as u64,
        );
        queue.submit(Some(encoder.finish()));

        let slice = point_out_staging.slice(..);
        let (sender, receiver) = oneshot::channel();

        slice.map_async(wgpu::MapMode::Read, move |result| {
            let _ = sender.send(result.unwrap());
        });

        device.poll(wgpu::Maintain::Wait);
        receiver.await.unwrap();

        let data = slice.get_mapped_range();
        let result: Vec<glam::Vec2> = bytemuck::cast_slice(&data).to_vec();
        // point_out_staging.unmap();

        result[0]
    }

    #[test]
    pub fn apply_linear_transform() {
        let transform = TransformSpec {
            coefs: Coefs::identity(),
            post_coefs: Coefs::identity(),
            variation_offset: 0,
            variation_count: 1,
        };
        let variations = vec![VariationSpec {
            kind: VariationKind::Linear,
            weight: 0.5,
            params: VariationParams::default(),
        }];
        let point_in = glam::vec2(1., 1.);

        let output_fut = do_apply_transform_f32(&transform, variations.as_slice(), &point_in);
        let result = block_on(output_fut);
        assert_eq!(result, glam::vec2(0.5, 0.5));

        // let mut point_out = glam::vec2(0., 0.);
        // flare_kernel::apply_transform_f32(glam::uvec3(0, 0, 0), &transform, &variations, &point_in, &mut point_out);
        // assert_eq!(point_out, glam::vec2(0.5, 0.5));
    }
}
