use serde::Deserialize;
use serde_xml_rs::from_reader;
use std::fs::File;
use std::io::BufReader;
use std::path::Path;
use thiserror::Error;
use xml::reader::XmlEvent;

#[derive(Debug, Default, Deserialize, Clone, PartialEq)]
pub struct Flames {
    pub name: String,
    #[serde(rename = "$value")]
    pub flames: Vec<Flame>,
}

#[derive(Debug, Default, Deserialize, Clone, PartialEq)]
pub struct Flame {
    pub name: String,
    pub version: String,
    pub palette: Option<u64>,
    #[serde(rename = "$value")]
    pub elements: Vec<FlameElement>,
}

#[derive(Debug, Deserialize, Clone, PartialEq)]
#[serde(rename_all = "lowercase")]
pub enum FlameElement {
    Xform(Xform),
    FinalXform(Xform),
    Color(Color),
    Palette(Palette),
    Edit(Edit),
}

#[derive(Debug, Default, Deserialize, Clone, PartialEq)]
pub struct Xform {
    // Weight is optional for final xform
    pub weight: Option<f64>,
    pub color: f64,
    pub coefs: String,

    pub linear: Option<f64>,
    pub julia: Option<f64>,
    pub popcorn: Option<f64>,
    pub pdj: Option<f64>,
    pub pdj_a: Option<f64>,
    pub pdj_b: Option<f64>,
    pub pdj_c: Option<f64>,
    pub pdj_d: Option<f64>,
}

#[derive(Debug, Default, Deserialize, Clone, PartialEq)]
pub struct Color {
    pub index: u64,
    pub rgb: String,
}

#[derive(Debug, Default, Deserialize, Clone, PartialEq)]
pub struct Palette {
    pub count: usize,
    pub format: String,
    #[serde(rename = "$value")]
    pub value: String,
}

#[derive(Debug, Deserialize, Default, Clone, PartialEq)]
pub struct Edit {}

#[derive(Error, Debug)]
pub enum ParseError {
    #[error("unable to read file; {0}")]
    IoError(#[from] std::io::Error),
    #[error("unable to parse file; {0}")]
    DeserializeError(#[from] serde_xml_rs::Error),
    #[error("unable to parse file")]
    ParseError,
    #[error("unable to parse xml; {0}")]
    XmlError(#[from] xml::reader::Error),
}

pub fn parse(p: impl AsRef<Path>) -> Result<Flames, ParseError> {
    let file = File::open(&p)?;
    let reader = BufReader::new(file);
    let mut xml_reader = xml::EventReader::new(reader);

    match xml_reader.next()? {
        XmlEvent::StartDocument { .. } => {}
        _ => return Err(ParseError::ParseError),
    }

    match xml_reader.next()? {
        XmlEvent::StartElement { name, .. } => {
            let file = File::open(&p)?;
            let reader = BufReader::new(file);
            match name.local_name.as_str() {
                "Flames" => from_reader(reader).map_err(|e| e.into()),
                "flame" => from_reader(reader)
                    .map(|f| Flames {
                        flames: vec![f],
                        ..Default::default()
                    })
                    .map_err(|e| e.into()),
                _ => Err(ParseError::ParseError),
            }
        }
        _ => Err(ParseError::ParseError),
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::path::PathBuf;

    fn resource_file(p: impl AsRef<Path>) -> PathBuf {
        PathBuf::from(env!("CARGO_MANIFEST_DIR"))
            .join("resources/test")
            .join(p)
    }

    #[test]
    fn param_247_54484() {
        let path = resource_file("247.54484.flame");
        let flames_result = parse(&path);

        assert!(
            matches!(flames_result, Ok(_)),
            "path = {} error = {}",
            path.as_path().display(),
            flames_result.unwrap_err()
        );
    }

    #[test]
    fn param_apophysis() {
        let path = resource_file("apophysis.flame");
        let flames_result = parse(&path);

        assert!(
            matches!(flames_result, Ok(_)),
            "path = {} error = {}",
            path.as_path().display(),
            flames_result.unwrap_err()
        );

        let flame = &flames_result.unwrap().flames[0];
        assert_eq!(flame.name, "post xform");

        let flame_element = &flame.elements[0];
        match flame_element {
            FlameElement::Xform(xform) => {
                assert_eq!(xform.weight, Some(0.422330042096567));
                assert_eq!(xform.pdj, Some(1.));
                assert_eq!(xform.pdj_a, Some(1.09358));
                assert_eq!(xform.pdj_b, Some(2.13048));
                assert_eq!(xform.pdj_c, Some(2.54127));
                assert_eq!(xform.pdj_d, Some(2.37267));
            }
            _ => assert!(false, "unexpected element"),
        }
    }
}
